:- module(
  biocham,
  [
    % Public API
    start/0,
    initialize/0,
    initial/1,
    biocham_command/0,
    biocham_command/1,
    library_path/1
  ]).


% Only for separate compilation/linting
:- use_module(doc).


:- doc('
The Biochemical Abstract Machine (Biocham) is a modelling software for cell systems biology, with some unique features for static and dynamic analyses using temporal logic constraints.
Biocham v4 is mainly composed of :
\\begin{itemize}
\\item 
a rule-based language for modeling biochemical processes with reactions and/or influences (respectively SBML2 and SBML3-qual compatible),
\\item
a hierarchy of semantics (differential, stochastic, Petri Net, Boolean) to interpret such models,
\\item
a temporal logic based language to formalize the temporal behaviors of biochemical systems,
\\item
several unique features for developing/analyzing/correcting/completing/calibrating/coupling/synthesizing reaction and influence systems with respect to formal specifications of their behaviour.
\\end{itemize}

Biocham v4 is a complete rewriting of Biocham v3.
It introduces some new features, mainly:
\\begin{itemize}
\\item influence network models,
\\item quantitative temporal logic patterns,
\\item compilation of functions in reaction networks,
\\item learning models from data.
\\end{itemize}
Some features of Biocham v3 are not implemented in v4.0, mainly:
\\begin{itemize}
\\item the Graphical User Interface (only the notebook)
\\item the graphical editor (SBGN compatible)
\\item the sensivity and robustness measures
\\item the detection of model reductions by subgraph epimorphisms (SEPI)
\\item hybrid simulations
\\end{itemize}
').

:- doc('
  Biocham is a free software protected by the GNU General Public License GPL
  version 2. This is an Open Source license that allows free usage of this
  software.
').

:- doc('This reference manual (and its extended version for developpers) is automaticaly generated from the source code of Biocham.').


start :-
  about,
  initialize,
  do_arguments.


:- dynamic(initialized/0).


initialize :-
  initialized,
  !.

initialize :-
  initialize_library_path,
  set_prolog_flag(allow_variable_name_as_functor, true),
  prolog_history(enable),
  set_counter(item_id, 0),
  set_plot_driver(gnu_plot),
  set_image_viewer_driver(open_file),
  set_draw_graph_driver(graph_pdf),
  nb_setval(graph_pdf, 0),
  nb_setval(current_models, []),
  load_biocham('library:initial'),
  \+ (
    initial_command(Command),
    \+ (
      command(Command)
    )
  ),
  new_model,
  assertz(initialized).


:- dynamic(initial_command/1).


initial(Command) :-
  devdoc('
    Adds a command to run during initialization
    (in the context of the "initial" model)
  '),
  assertz(initial_command(Command)).


:- dynamic(library_path/1).


initialize_library_path :-
  (
    getenv('BIOCHAMLIB', LibraryPath)
  ->
    true
  ;
    executable_filename(ExecutableFilename),
    absolute_file_name(
      'library/gsl_solver.c',
      LibraryFile,
      [relative_to(ExecutableFilename)]
    ),
    file_directory_name(LibraryFile, LibraryPath)
  ),
  assertz(library_path(LibraryPath)).


biocham_command :-
  doc('Declares a command for the top-level.').


biocham_command(_) :-
  doc('
    With a * : declares a command for the top-level that can take an
    unlimited number of arguments.
    The additional arguments are given as a list in the last parameter.').

:-devcom('\\begin{todo}Make biocham commands functional. We need to know the result type for the gui... Furthermore, the show, list, enumerate,... commands are not so nice. The print result should be a text or a widget associated to the returned type.\\end{todo}').

